<?php

/**
 * Field handler for displaying Solr fields in Views.
 */
class SarniaViewsHandlerField extends views_handler_field {

  /**
   * A Sarnia entity bundle info array.
   * @var array
   */
  protected $sarnia_type_info;

  /**
   * Initialize the handler and load Sarnia entity type info.
   */
  function init(&$view, &$options) {
    parent::init($view, $options);
    $this->sarnia_type_info = sarnia_entity_type_load_by_index($this->definition['search_api_index']);
  }

  /**
   * Provide a custom UI name for the field based on the Solr property.
   */
  function ui_name($short = FALSE) {
    if (!empty($this->options['ui_name'])) {
      $title = check_plain($this->options['ui_name']);
      return $title;
    }

    if (isset($this->options['solr_property'])) {
      $title = check_plain($this->options['solr_property']);
    }
    else {
      $title = ($short && isset($this->definition['title short'])) ? $this->definition['title short'] : $this->definition['title'];
    }

    return t('!group: !title', array('!group' => $this->definition['group'], '!title' => $title));
  }

  /**
   * Specify handler options, adding formatter-specific properties.
   */
  public function option_definition() {
    $options = parent::option_definition();

    $options['solr_property'] = array('default' => NULL);
    $options['formatter'] = array('default' => NULL);
    $options['settings'] = array('default' => array());

    return $options;
  }

  /**
   * Build the handler configuration form.
   *
   * Adds a Solr property select dropdown, and summons up a formatter settings
   * form.
   *
   * @see views_handler_field_field::options_form()
   */
  public function options_form(&$form, &$form_state) {
    parent::options_form($form, $form_state);

    $form['solr_property'] = array(
      '#type' => 'select',
      '#title' => t('Solr property'),
      '#options' => sarnia_index_get_field_options($this->definition['search_api_index']),
      '#default_value' => $this->options['solr_property'],
    );
    sarnia_element_add_combobox($form['solr_property']);

    // From views/modules/field/views_handler_field_field.inc
    module_load_include('inc', 'views', 'modules/field/views_handler_field_field');
    $field = field_info_field($this->definition['field_name']);
    $formatters = array(t('- None -')) + _field_view_formatter_options($field['type']);
    $form['formatter'] = array(
      '#type' => 'select',
      '#title' => t('Formatter'),
      '#options' => $formatters,
      '#default_value' => $this->options['formatter'],
      '#ajax' => array(
        'path' => views_ui_build_form_url($form_state),
      ),
      '#submit' => array('views_ui_config_item_form_submit_temporary'),
      '#executes_submit_callback' => TRUE,
    );

    // Get the currently selected formatter.
    $format = $this->options['formatter'];
    $formatter = field_info_formatter_types($format);
    $settings = $this->options['settings'] + field_info_formatter_settings($format);

    // Provide an instance array for hook_field_formatter_settings_form(). This
    // mimics ctools_fields_fake_field_instance(), except it gets a real
    // instance since we know our entity type.
    $entity_type = $this->sarnia_type_info['machine_name'];
    $instance = field_info_instance($entity_type, $this->definition['field_name'], $entity_type);
    $instance['display']['_dummy'] = array(
      'type' => $formatter,
      'settings' => $settings,
    );

    // Store the settings in a '_dummy' view mode.
    $instance['display']['_dummy'] = array(
      'type' => $format,
      'settings' => $settings,
    );

    // Get the settings form.
    $settings_form = array('#value' => array());
    $function = (isset($formatter['module']) ? $formatter['module'] : '') . '_field_formatter_settings_form';
    if (function_exists($function)) {
      // views_handler_field_field.inc passes $form instead of array()
      $settings_form = $function($field, $instance, '_dummy', array(), $form_state);
      unset($settings_form['solr_property']);
    }
    $form['settings'] = $settings_form;
  }

  function options_submit(&$form, &$form_state) {
    parent::options_submit($form, $form_state);

    // Set a reasonable label if it is the first save and none was specified, or
    // if the the previous label matched the previous property name and has not
    // been changed.
    if ((!isset($this->options['solr_property']) && $this->options['label'] == $this->definition['title']) ||
        ($this->options['solr_property'] == $this->options['label'] && $this->options['label'] == $form_state['values']['options']['label'])) {
      $form_state['values']['options']['label'] = $form_state['values']['options']['solr_property'];
    }
  }

  /**
   * Add this handler's data to the query.
   *
   * By default Solr returns all of the fields in a Solr doc. All that this
   * method needs to do is specify the field alias.
   */
  public function query() {
    $this->field_alias = $this->options['solr_property'];
  }

  /**
   * Load any entities that aren't present.
   */
  function pre_render(&$results) {
    // Don't bother trying to load entities if the formatter doesn't need them.
    if (empty($this->options['formatter'])) {
      return;
    }

    // Get a list of unloaded entities.
    $ids = array();
    foreach ($results as $result_id => $result) {
      $entity_id = &$result->entity;
      if (!is_object($entity_id)) {
        $ids[$result_id] = $entity_id;
      }
    }
    // Load whatever unloaded entities we found.
    if (!empty($ids)) {
      $entities = entity_load($this->sarnia_type_info['machine_name'], array_values($ids));
      foreach ($ids as $result_id => $entity_id) {
        if ($entities[$entity_id]) {
          $results[$result_id]->entity = $entities[$entity_id];
        }
      }
    }
  }

  /**
   * Render a field value.
   */
  function render($values) {
    $value = $this->get_value($values);
    $output = '';

    // Is this formatted using a Field API field formatter? Use the same check
    // as in SarniaViewsHandlerField::get_value().
    $id_field = $this->sarnia_type_info['id_field'];
    if (!empty($this->options['formatter']) && !empty($values->entity)) {
      foreach (element_children($value) as $key) {
        $output .= $value[$key]['rendered'];
      }
    }
    elseif (is_array($value)) {
      // Plain multivalue results.
      $list = array(
        '#theme' => 'item_list',
        '#items' => array(),
      );
      foreach ($value as $v) {
        $list['#items'][] = $this->sanitize_value($v);
      }
      $output = drupal_render($list);
    }
    else {
      // Plain single value results.
      $output = $this->sanitize_value($value);
    }

    return $output;
  }

  /**
   * Get the value to render.
   *
   * @param $values
   *   An object containing all retrieved values.
   * @param $field
   *   Optional name of the field where the value is stored.
   */
  public function get_value($values, $field = NULL) {
    // Are we using a Field API field formatter, and is our entity loaded in the results?
    $id_field = $this->sarnia_type_info['id_field'];
    if (!empty($this->options['formatter']) && !empty($values->entity)) {
      // Build arguments for the Field API field formatter.
      $field_name = $this->definition['field_name'];
      $display = array(
        'type' => $this->options['formatter'],
        'settings' => $this->options['settings'] + array('solr_property' => $this->options['solr_property']),
        'label' => 'hidden',
      );
      $render_array = field_view_field($values->entity->type, $values->entity, $field_name, $display, LANGUAGE_NONE);

      // Is this handler configured to use Field API markup for fields?
      if (!empty($this->options['field_api_classes'])) {
        // Return items wrapped in Field API field markup.
        return array(array(
          'rendered' => drupal_render($render_array),
          'raw' => $render_array['#items'],
        ));
      }

      // Strip extra details out of the render array so as to return plain items.
      $items = array();
      foreach (element_children($render_array) as $key) {
        $items[$key]['rendered'] = drupal_render($render_array[$key]);
        $items[$key]['raw'] = $render_array['#items'][$key];
      }
      return $items;
    }
    else {
      // Return an unprocessed value.
      return parent::get_value($values, $field);
    }
  }

  /**
   * Re-title tokens from Sarnia field handlers with their Solr property names.
   */
  function document_self_tokens(&$tokens) {
    foreach ($this->view->display_handler->get_handlers('field') as $field => $handler) {
      if (is_a($handler, 'SarniaViewsHandlerField')) {
        $token_name = "[$field]";
        if (isset($tokens[$token_name]) && isset($handler->options['solr_property'])) {
          $tokens[$token_name] = $handler->admin_summary();
        }
      }
    }
  }

  /**
   * Override the default administration interface summary to include the Solr
   * property name.
   */
  public function admin_summary() {
    if (strtolower($this->options['solr_property']) != strtolower($this->options['label'])) {
      $summary = check_plain($this->options['label']) . ' [' . check_plain($this->options['solr_property']) . ']';
    }
    else {
      $summary = $this->options['solr_property'];
    }
    return $summary;
  }

  /**
   * Check whether this field is sortable.
   */
  function click_sortable() {
    $name = $this->options['solr_property'];
    $server = search_api_server_load($this->sarnia_type_info['search_api_server']);
    $fields = $server->getSortFields();
    return isset($fields[$name]);
  }

  /**
   * Add a click sort to the query.
   */
  public function click_sort($order) {
    $this->query->sort($this->field_alias, $order);
  }

}
